// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include <iostream>
#include <dune/common/parallel/mpihelper.hh> // An initializer of MPI
#include <dune/common/exceptions.hh> // We use exceptions
#include <dune/alugrid/grid.hh>
#include <dune/curvedsurfacegrid/curvedsurfacegrid.hh>
#include <dune/grid/io/file/gmshreader.hh>
#include <dune/grid/io/file/vtk/vtkwriter.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>

using namespace Dune;

int main(int argc, char** argv)
{
  try{
    // Maybe initialize MPI
    Dune::MPIHelper& helper = Dune::MPIHelper::instance(argc, argv);
    std::cout << "Hello World! This is an example for dune-curvedsurfacegrid." << std::endl;
    if(Dune::MPIHelper::isFake)
      std::cout<< "This is a sequential program." << std::endl;
    else
      std::cout<<"I am rank "<<helper.rank()<<" of "<<helper.size()
        <<" processes!"<<std::endl;
  }
  catch (Dune::Exception &e){
    std::cerr << "Dune reported error: " << e << std::endl;
  }
  catch (...){
    std::cerr << "Unknown exception thrown!" << std::endl;
  }

  //load host-grid (ALUGrid)
  using HostGridType = Dune::ALUGrid<2, 3, simplex, conforming>;
  GridFactory<HostGridType> gridFactory;
  GmshReader<HostGridType>::read(gridFactory, "example/sphere_coarse.msh");
  std::shared_ptr<HostGridType> hostGrid(gridFactory.createGrid());

  const int ip_order = 3; //interpolatory order for curved elements

  //case1: projection to actual sphere
  {
    SphereCoordFunction sphereCF(/*center=*/{0.0, 0.0, 0.0}, /*radius=*/1.0);
    CurvedSurfaceGrid grid(*hostGrid, sphereCF, std::integral_constant<int, ip_order>{});

    auto gridView = grid.leafGridView();
    using GridView = decltype(gridView);
    VTKWriter<GridView> vtkWriter(gridView);
    vtkWriter.write("sphere1_real_elements");
    SubsamplingVTKWriter<GridView> ssvtkWriter(gridView, refinementIntervals(ip_order));
    ssvtkWriter.write("sphere1_fake_elements" + CGeo::ft(ip_order, 2));
  }

  //case2: projection to a surface-mesh of a sphere
  {
    SurfaceDistanceCoordFunction<> surfdistCF("example/sphere_fine.vtu");
    CurvedSurfaceGrid grid(*hostGrid, std::ref(surfdistCF), std::integral_constant<int, ip_order>{});
    
    auto gridView = grid.leafGridView();
    using GridView = decltype(gridView);
    VTKWriter<GridView> vtkWriter(gridView);
    vtkWriter.write("sphere2_real_elements");
    SubsamplingVTKWriter<GridView> ssvtkWriter(gridView, refinementIntervals(ip_order));
    ssvtkWriter.write("sphere2_fake_elements" + CGeo::ft(ip_order, 2));
  }

  return 0;
}
